#ifndef HIBDRV_H
#define HIBDRV_H

#include <asm/page.h>
#include <linux/ioctl.h>
#include "typedef.h"

#define HIBDRV_DEVNAME "hibdrv"
#define NHIBDRV (0) /* max # of device */

#define HIBDRV_VENDORID (0x1b1a)
#define HIBDRV_DEVICEID (0x0e70)

// size of DMA buffers
#ifndef PAGE_SIZE
#define PAGE_SIZE sysconf (_SC_PAGE_SIZE)
#endif
#define HIBDRV_DMABUF_ORDER (5) /* 2^5 pages */
#define HIBDRV_DMABUF_BYTES (PAGE_SIZE * (1<<HIBDRV_DMABUF_ORDER)) /* 128kbyte, 32pages */
#define HIBDRV_DMABUF_WORDS (HIBDRV_DMABUF_BYTES>>2) /* dma buffer size in quad-word */

#define HIBDRV_PIOWBUF_BYTES (4096)     // size of double buffer for PIO write
#define HIBDRV_LOCAL_MEM_BYTES (8192)   // size of local register
#define HIBDRV_BACKEND_MEM_BYTES (8192) // size of backend space

/* structure for read/write ioctl */

enum {
    HIBDRV_MAP_HIBMEM = 0,
    HIBDRV_MAP_DMARBUF,
    HIBDRV_MAP_DMAWBUF,
    HIBDRV_MAP_BACKEND,
    HIBDRV_MAP_DBLBUF,
};

#define HIBDRV_MAGIC (5)
#define HIBDRV_WRITE_CFG            _IOW(HIBDRV_MAGIC, 5, struct long_access)
#define HIBDRV_READ_CFG             _IOWR(HIBDRV_MAGIC, 4, struct long_access)
#define HIBDRV_GET_DMAR_PA          _IOWR(HIBDRV_MAGIC, 18, unsigned long)
#define HIBDRV_GET_DMAW_PA          _IOWR(HIBDRV_MAGIC, 19, unsigned long)
#define HIBDRV_SET_MAP_MODE         _IOW(HIBDRV_MAGIC, 20, unsigned int)


/*
 *
 * hib local registers including
 * DMA registers are mapped to BAR0.
 *
 * DMA channel0 for read (host->hib),
 * DMA channel1  for write (host<-hib).
 *
 * register map:
 *
 * 000h DMA0  lower 32bits of start address in PCI space.
 * 004h DMA0  higer 32bits of start address in PCI space.
 * 008h DMA0  start address in hib local space.
 * 00ch DMA1  transfer size in byte.
 * 010h DMA0  command/parameter
 *              0      : 0:fifo mode  1:ram mode
 *              1      : not used.
 *              3:2    : read latency. 00:1clk  01:2clk  10:3clk  11:4clk
 *              7:4    : be for DW transfer. not used for burst transfer.
 *              8      : direction. '0' for MemRd burst. '1' for MemWr burst.
 *              10:9   : not used.
 *              13:11  : TC. tied to "000".
 *              15:14  : ATTR. tied to "00".
 *              16     : kick off a transfer.
 *              17     : abort current transfer.
 * 014h DMA0  status
 *              0      : asserted when current transfer is done.
 * 040h DMA1  lower 32bits of start address in PCI space.
 * 044h DMA1  higer 32bits of start address in PCI space.
 * 048h DMA1  start address in hib local space.
 * 04ch DMA1  transfer size in byte, write to this register starts a new transfer.
 * 050h DMA1  command/parameter
 * 054h DMA1  status
 *
 * 120h DMA status
 * 128h bord function test mode, piow done.
 * 130h KFCR board info
 * 138h ASMI
 * 140h PLLCONF
 */

#define HIBDRV_DMA0LSB   (0x000)
#define HIBDRV_DMA0MSB   (0x004)
#define HIBDRV_DMA0LADDR (0x008)
#define HIBDRV_DMA0SIZE  (0x00c)
#define HIBDRV_DMA0CMD   (0x010)
#define HIBDRV_DMA0MISC  (0x014)

#define HIBDRV_DMA1LSB   (0x048)
#define HIBDRV_DMA1MSB   (0x044)
#define HIBDRV_DMA1LADDR (0x040)
#define HIBDRV_DMA1SIZE  (0x04c)
#define HIBDRV_DMA1CMD   (0x050)
#define HIBDRV_DMA1MISC  (0x054)

#define HIBDRV_COMMAND    (0x120)
#define HIBDRV_TESTMODE   (0x128)
#define HIBDRV_BOARDINFO  (0x130)
#define HIBDRV_ASMI       (0x138)
#define HIBDRV_PLLCONF    (0x140)

#define HIBDRV_DMA0CMD_KICKOFF_BIT   (0)
#define HIBDRV_DMA0CMD_ABORT_BIT     (1)
#define HIBDRV_DMA0MISC_DONE_BIT     (13)

#define HIBDRV_DMA1CMD_KICKOFF_BIT   (0)
#define HIBDRV_DMA1CMD_ABORT_BIT     (1)
#define HIBDRV_DMA1MISC_DONE_BIT     (13)

#define HIBDRV_COMMAND_RESET_BACKEND_BIT (31)
#define HIBDRV_COMMAND_SRAM1_WLOCK_BIT   (26)
#define HIBDRV_COMMAND_SRAM0_WLOCK_BIT   (25)
#define HIBDRV_COMMAND_SWAP_SRAM_BIT     (24)
#define HIBDRV_COMMAND_SRAM_WCNT_BIT     (0)
/* command register
 * 31               : reset backend.                           (wo)
 * 26               : piow buf1 is write locked.               (ro)
 * 25               : piow buf0 is write locked.               (ro)
 * 24               : swap the two piow buffers.               (wo)
 * PIOWBUF_DEPTH..0 : piow buf0/1 write count (in byte unit).  (r/w)
 */

#define HIBDRV_ASMI_ERROR_BIT (1)
#define HIBDRV_ASMI_BUSY_BIT  (0)
#define HIBDRV_ASMI_CMD_RSID  (0xfffffd00)
#define HIBDRV_ASMI_CMD_WP    (0xfffffe00)
#define HIBDRV_ASMI_CMD_EB    (0xffffff00)

#define HIBDRV_PLLCONF_WRITE_BIT    (0)
#define HIBDRV_PLLCONF_RECONFIG_BIT (2)
#define HIBDRV_PLLCONF_BUSY_BIT     (0)

#endif /* HIBDRV_H */
