#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <assert.h>

#include <xmmintrin.h>
#include <emmintrin.h>

#include "hibutil.h"

#define NDMAWSEQ 4 // number of DMAW sequences
#define NDMASEQ (NDMAWSEQ+NDMAWSEQ)


/* HIB local definition(s) & variable(s) */
Hib hib[NHIB];
int hib_fd[NHIB];
unsigned long dmar_pa[NHIB]; /* physical addres of hib[].dmar_buf */
unsigned long dmaw_pa[NHIB]; /* physical addres of hib[].dmaw_buf */

// HIB extention DMA registers added by JM
#define HIB_DMA_ADRREG (0x60)
#define HIB_DMA_DATAREG (0x64)
#define HIB_DMA_TRIGREG (0x68)

typedef struct{
    UINT64* buf;
    int size;
}DMA_SEQ,*PDMA_SEQ;

DMA_SEQ dmaw_seq[NHIB][NDMAWSEQ];
DMA_SEQ dmar_seq[NHIB][NDMAWSEQ];

static void (*SendMCfunc[NHIB])(int devid, int size, UINT64 *buf);
static int (*RecvMCfunc[NHIB])(int devid, int size, UINT64 *buf);

/* HIB local function(s) */
static Hib* sub_openMC(int devid);
static void init_envs(int devid);
static void lock_hib(int devid);
static int try_lock_hib(int devid);
static void unlock_hib(int devid);
static void TBmemMap(int devid);
static void TBgetDmaInfo(int devid,
			 unsigned long *dmarpa, UINT64 **dmarva,
			 unsigned long *dmawpa, UINT64 **dmawva);
static void hib_config_setmaxrequestMC0(int devid);


#define WARN(lv, fmt, args...) if (lv <= warn_level) fprintf(stderr, fmt, ## args);
static int warn_level = 2; /* warning message output level. the higher the more verbose.
                              0: no warning (may cause wrong result with ../scripts/check.csh)
                              1: minimum
                              2: default
                              3: for debugging purpose
                           */

void
hib_set_warn_level(int lv)
{
    warn_level = lv;
}

int
hib_recvMC(int devid, int size, UINT64 *buf)
{
    return RecvMCfunc[devid](devid, size, buf);
}

/*
 * open HIB
 * returns a ponter to the Hib object opened.
 */
Hib*
hib_openMC(int devid)
{
    init_envs(devid);

    //    fprintf(stderr,"hib_openMC, devid=%d name=%s\n", devid, hib[devid].devname);

    hib_fd[devid] = open(hib[devid].devname, O_RDWR);
    if (hib_fd[devid] < 0) {
	perror(hib[devid].devname);
	exit(2);
    }

    lock_hib(devid);
    return sub_openMC(devid);
} 

/*
 * try to open HIB.
 * mostly the same as hib_openMC(), but this version does not block
 * when the device is already opend by another process.
 * returns a ponter to the Hib object if open succeeded, otherwise returns NULL.
 */
Hib*
hib_try_openMC(int devid)
{
    init_envs(devid);

    hib_fd[devid] = open(hib[devid].devname, O_RDWR);
    if (hib_fd[devid] < 0) {
	perror(hib[devid].devname);
	exit(2);
    }

    if (try_lock_hib(devid) < 0) { // someone is using this device. giving up.
        return NULL;
    }
    return sub_openMC(devid);
}

static Hib* sub_openMC(int devid)
{
    /*
     * set local register addresses
     */
    switch(hib[devid].type) {

      case HIB_GRAPE7X:
        hib[devid].dmabuf_bytes = GRAPE7X_DMABUF_BYTES;
        hib[devid].piowbuf_bytes = GRAPE7X_PIOWBUF_BYTES;
        hib[devid].local_mem_bytes = GRAPE7X_LOCAL_MEM_BYTES;
        hib[devid].backend_mem_bytes = GRAPE7X_BACKEND_MEM_BYTES;

        hib[devid].write_cfg = GRAPE7X_WRITE_CFG;
        hib[devid].read_cfg = GRAPE7X_READ_CFG;
        hib[devid].get_dmaw_pa = GRAPE7X_GET_DMAW_PA;
        hib[devid].get_dmar_pa = GRAPE7X_GET_DMAR_PA;
        hib[devid].set_map_mode = GRAPE7X_SET_MAP_MODE;

        hib[devid].map_hibmem = GRAPE7X_MAP_HIBMEM;
        hib[devid].map_dmarbuf = GRAPE7X_MAP_DMARBUF;
        hib[devid].map_dmawbuf = GRAPE7X_MAP_DMAWBUF;
        hib[devid].map_backend = GRAPE7X_MAP_BACKEND;
        hib[devid].map_dblbuf = GRAPE7X_MAP_DBLBUF;

        hib[devid].boardinfo = GRAPE7X_BOARDINFO;
        hib[devid].dma0lsb = GRAPE7X_DMA0LSB;
        hib[devid].dma0msb = GRAPE7X_DMA0MSB;
        hib[devid].dma0size = GRAPE7X_DMA0SIZE;
        hib[devid].dma0cmd = GRAPE7X_DMA0CMD;
        hib[devid].dma0misc = GRAPE7X_DMAMISC;
        hib[devid].dma1lsb = GRAPE7X_DMA1LSB;
        hib[devid].dma1msb = GRAPE7X_DMA1MSB;
        hib[devid].dma1size = GRAPE7X_DMA1SIZE;
        hib[devid].dma1cmd = GRAPE7X_DMA1CMD;
        hib[devid].dma1misc = GRAPE7X_DMAMISC;
        hib[devid].command = GRAPE7X_COMMAND;
        hib[devid].testmode = GRAPE7X_TESTMODE;
        hib[devid].asmi = GRAPE7X_ASMI;
        hib[devid].pllconf = GRAPE7X_PLLCONF;

        hib[devid].dma0cmd_kickoff_bit       = GRAPE7X_DMA0CMD_KICKOFF_BIT;
        hib[devid].dma0misc_done_bit         = GRAPE7X_DMA0MISC_DONE_BIT;
        hib[devid].dma0misc_swap_sram_bit    = GRAPE7X_DMA0MISC_SWAP_SRAM_BIT;
        hib[devid].dma0misc_sram_wcnt_bit    = GRAPE7X_DMA0MISC_SRAM_WCNT_BIT;
        hib[devid].dma1cmd_kickoff_bit       = GRAPE7X_DMA1CMD_KICKOFF_BIT;
        hib[devid].dma1misc_done_bit         = GRAPE7X_DMA1MISC_DONE_BIT;
        hib[devid].command_reset_backend_bit = GRAPE7X_COMMAND_RESET_BACKEND_BIT;
        hib[devid].command_dma_reset_bit     = GRAPE7X_COMMAND_DMA_RESET_BIT;
        hib[devid].pllconf_write_bit         = GRAPE7X_PLLCONF_WRITE_BIT;
        hib[devid].pllconf_reconfig_bit      = GRAPE7X_PLLCONF_RECONFIG_BIT;
        hib[devid].pllconf_busy_bit          = GRAPE7X_PLLCONF_BUSY_BIT;

        hib[devid].asmi_error_bit = GRAPE7X_ASMI_ERROR_BIT;
        hib[devid].asmi_busy_bit  = GRAPE7X_ASMI_BUSY_BIT;
        hib[devid].asmi_cmd_rsid  = GRAPE7X_ASMI_CMD_RSID;
        hib[devid].asmi_cmd_wp    = GRAPE7X_ASMI_CMD_WP;
        hib[devid].asmi_cmd_eb    = GRAPE7X_ASMI_CMD_EB;

        break;

      case HIB_GRAPE7E:
        hib[devid].dmabuf_bytes = GRAPE7E_DMABUF_BYTES;
        hib[devid].piowbuf_bytes = GRAPE7E_PIOWBUF_BYTES;
        hib[devid].local_mem_bytes = GRAPE7E_LOCAL_MEM_BYTES;
        hib[devid].backend_mem_bytes = GRAPE7E_BACKEND_MEM_BYTES;

        hib[devid].write_cfg = GRAPE7E_WRITE_CFG;
        hib[devid].read_cfg = GRAPE7E_READ_CFG;
        hib[devid].get_dmaw_pa = GRAPE7E_GET_DMAW_PA;
        hib[devid].get_dmar_pa = GRAPE7E_GET_DMAR_PA;
        hib[devid].set_map_mode = GRAPE7E_SET_MAP_MODE;

        hib[devid].map_hibmem = GRAPE7E_MAP_HIBMEM;
        hib[devid].map_dmarbuf = GRAPE7E_MAP_DMARBUF;
        hib[devid].map_dmawbuf = GRAPE7E_MAP_DMAWBUF;
        hib[devid].map_backend = GRAPE7E_MAP_BACKEND;
        hib[devid].map_dblbuf = GRAPE7E_MAP_DBLBUF;

        hib[devid].boardinfo = GRAPE7E_BOARDINFO;
        hib[devid].dma0lsb = GRAPE7E_DMA0LSB;
        hib[devid].dma0msb = GRAPE7E_DMA0MSB;
        hib[devid].dma0size = GRAPE7E_DMA0SIZE;
        hib[devid].dma0cmd = GRAPE7E_DMA0CMD;
        hib[devid].dma0misc = GRAPE7E_DMA0MISC;
        hib[devid].dma1lsb = GRAPE7E_DMA1LSB;
        hib[devid].dma1msb = GRAPE7E_DMA1MSB;
        hib[devid].dma1size = GRAPE7E_DMA1SIZE;
        hib[devid].dma1cmd = GRAPE7E_DMA1CMD;
        hib[devid].dma1misc = GRAPE7E_DMA1MISC;
        hib[devid].command = GRAPE7E_COMMAND;
        hib[devid].testmode = GRAPE7E_TESTMODE;
        hib[devid].asmi = GRAPE7E_ASMI;
        hib[devid].pllconf = GRAPE7E_PLLCONF;

        hib[devid].dma0cmd_kickoff_bit       = GRAPE7E_DMA0CMD_KICKOFF_BIT;
        hib[devid].dma0misc_done_bit         = GRAPE7E_DMA0MISC_DONE_BIT;
        hib[devid].dma0misc_swap_sram_bit    = GRAPE7E_DMA0MISC_SWAP_SRAM_BIT;
        hib[devid].dma0misc_sram_wcnt_bit    = GRAPE7E_DMA0MISC_SRAM_WCNT_BIT;
        hib[devid].dma1cmd_kickoff_bit       = GRAPE7E_DMA1CMD_KICKOFF_BIT;
        hib[devid].dma1misc_done_bit         = GRAPE7E_DMA1MISC_DONE_BIT;
        hib[devid].command_reset_backend_bit = GRAPE7E_COMMAND_RESET_BACKEND_BIT;
        hib[devid].command_dma_reset_bit     = GRAPE7E_COMMAND_DMA_RESET_BIT;
        hib[devid].pllconf_write_bit         = GRAPE7E_PLLCONF_WRITE_BIT;
        hib[devid].pllconf_reconfig_bit      = GRAPE7E_PLLCONF_RECONFIG_BIT;
        hib[devid].pllconf_busy_bit          = GRAPE7E_PLLCONF_BUSY_BIT;

        hib[devid].asmi_error_bit = GRAPE7E_ASMI_ERROR_BIT;
        hib[devid].asmi_busy_bit  = GRAPE7E_ASMI_BUSY_BIT;
        hib[devid].asmi_cmd_rsid  = GRAPE7E_ASMI_CMD_RSID;
        hib[devid].asmi_cmd_wp    = GRAPE7E_ASMI_CMD_WP;
        hib[devid].asmi_cmd_eb    = GRAPE7E_ASMI_CMD_EB;

        break;

      case HIB_HIBDRV:
        hib[devid].dmabuf_bytes = HIBDRV_DMABUF_BYTES;
        hib[devid].piowbuf_bytes = HIBDRV_PIOWBUF_BYTES;
        hib[devid].local_mem_bytes = HIBDRV_LOCAL_MEM_BYTES;
        hib[devid].backend_mem_bytes = HIBDRV_BACKEND_MEM_BYTES;

        hib[devid].write_cfg = HIBDRV_WRITE_CFG;
        hib[devid].read_cfg = HIBDRV_READ_CFG;
        hib[devid].get_dmaw_pa = HIBDRV_GET_DMAW_PA;
        hib[devid].get_dmar_pa = HIBDRV_GET_DMAR_PA;
        hib[devid].set_map_mode = HIBDRV_SET_MAP_MODE;

        hib[devid].map_hibmem = HIBDRV_MAP_HIBMEM;
        hib[devid].map_dmarbuf = HIBDRV_MAP_DMARBUF;
        hib[devid].map_dmawbuf = HIBDRV_MAP_DMAWBUF;
        hib[devid].map_backend = HIBDRV_MAP_BACKEND;
        hib[devid].map_dblbuf = HIBDRV_MAP_DBLBUF;

        hib[devid].boardinfo = HIBDRV_BOARDINFO;
        hib[devid].dma0lsb = HIBDRV_DMA0LSB;
        hib[devid].dma0msb = HIBDRV_DMA0MSB;
        hib[devid].dma0laddr = HIBDRV_DMA0LADDR;
        hib[devid].dma0size = HIBDRV_DMA0SIZE;
        hib[devid].dma0cmd = HIBDRV_DMA0CMD;
        hib[devid].dma0misc = HIBDRV_DMA0MISC;
        hib[devid].dma1lsb = HIBDRV_DMA1LSB;
        hib[devid].dma1msb = HIBDRV_DMA1MSB;
        hib[devid].dma1laddr = HIBDRV_DMA1LADDR;
        hib[devid].dma1size = HIBDRV_DMA1SIZE;
        hib[devid].dma1cmd = HIBDRV_DMA1CMD;
        hib[devid].dma1misc = HIBDRV_DMA1MISC;
        hib[devid].command = HIBDRV_COMMAND;
        hib[devid].testmode = HIBDRV_TESTMODE;
        hib[devid].asmi = HIBDRV_ASMI;
        hib[devid].pllconf = HIBDRV_PLLCONF;

        hib[devid].dma0cmd_kickoff_bit       = HIBDRV_DMA0CMD_KICKOFF_BIT;
        hib[devid].dma0cmd_abort_bit         = HIBDRV_DMA0CMD_ABORT_BIT;
        hib[devid].dma0misc_done_bit         = HIBDRV_DMA0MISC_DONE_BIT;

        hib[devid].command_swap_sram_bit     = HIBDRV_COMMAND_SWAP_SRAM_BIT;
        hib[devid].command_sram_wcnt_bit     = HIBDRV_COMMAND_SRAM_WCNT_BIT;
        hib[devid].command_sram0_wlock_bit   = HIBDRV_COMMAND_SRAM0_WLOCK_BIT;
        hib[devid].command_sram1_wlock_bit   = HIBDRV_COMMAND_SRAM1_WLOCK_BIT;

        hib[devid].dma1cmd_kickoff_bit       = HIBDRV_DMA1CMD_KICKOFF_BIT;
        hib[devid].dma1cmd_abort_bit         = HIBDRV_DMA1CMD_ABORT_BIT;
        hib[devid].dma1misc_done_bit         = HIBDRV_DMA1MISC_DONE_BIT;
        hib[devid].command_reset_backend_bit = HIBDRV_COMMAND_RESET_BACKEND_BIT;
        hib[devid].pllconf_write_bit         = HIBDRV_PLLCONF_WRITE_BIT;
        hib[devid].pllconf_reconfig_bit      = HIBDRV_PLLCONF_RECONFIG_BIT;
        hib[devid].pllconf_busy_bit          = HIBDRV_PLLCONF_BUSY_BIT;

        hib[devid].asmi_error_bit = HIBDRV_ASMI_ERROR_BIT;
        hib[devid].asmi_busy_bit  = HIBDRV_ASMI_BUSY_BIT;
        hib[devid].asmi_cmd_rsid  = HIBDRV_ASMI_CMD_RSID;
        hib[devid].asmi_cmd_wp    = HIBDRV_ASMI_CMD_WP;
        hib[devid].asmi_cmd_eb    = HIBDRV_ASMI_CMD_EB;

        break;

      default:
        fprintf(stderr, "hib_openMC: unknown HIB type.\n", hib[devid].type);
        exit(1);
    }

#if 0
#warning this version of hib_openMC does not initialize memory map nor does DMA engine.
#else

    TBmemMap(devid);

    /*
     * maps DMA buffers
     */ 
    TBgetDmaInfo(devid,
		 &(dmar_pa[devid]), &(hib[devid].dmar_buf),
		 &(dmaw_pa[devid]), &(hib[devid].dmaw_buf));
    /*
     * dmar/w_buf: virtual address of DMA read/write buffer
     * dmar/w_pa: its physical address
     */
    WARN(3, "hib[%d].dmar_buf: 0x%016lx pa: %016lx\n",
         devid, hib[devid].dmar_buf, dmar_pa[devid]);
    WARN(3, "hib[%d].dmaw_buf: 0x%016lx pa: %016lx\n",
         devid, hib[devid].dmaw_buf, dmaw_pa[devid]);

    /* initialize local registers */
    hib_mem_writeMC(devid, hib[devid].dma0lsb, dmar_pa[devid]); /* LSB of DMA0 start address in PCI space */
    hib_mem_writeMC(devid, hib[devid].dma1lsb, dmaw_pa[devid]); /* LSB of DMA1 start address in PCI space */
    hib_mem_writeMC(devid, hib[devid].dma0msb, 0x0);
    hib_mem_writeMC(devid, hib[devid].dma1msb, 0x0);
    if (hib[devid].type == HIB_HIBDRV) {
        hib_mem_writeMC(devid, hib[devid].dma0laddr, 0x0); /* start address in local space */
        hib_mem_writeMC(devid, hib[devid].dma1laddr, 0x0); /* start address in local space */
        hib_mem_writeMC(devid, hib[devid].dma0cmd, 0); // set default param
        hib_mem_writeMC(devid, hib[devid].dma1cmd, 0); // set default param
    }

    hib_set_test_modeMC(devid, TESTMODE_NONE);

    SendMCfunc[devid] = hib_dmarMC;
    RecvMCfunc[devid] = hib_dmawMC;

#endif

       WARN(3, "hib[%d] opened.\n", devid);

    hib_config_setmaxrequestMC0(devid);
    usleep(10);
    hib_set_test_modeMC(devid, TESTMODE_NONE);
    usleep(10);
    hib_set_dma_extended_regMC(devid, 0, 0x0, 0x01f);
    usleep(10);
    hib_reset_dma(devid);
    usleep(10);
    
    return &(hib[devid]);
}

/* close HIB */
void
hib_closeMC(int devid)
{
    size_t psz = getpagesize();

    //    unlock_hib(devid);

#if 0
#warning this version of hib_closeMC does not clean up memory map nor does DMA engine.
#else
    munmap((void *)hib[devid].local_reg,   (hib[devid].local_mem_bytes/psz+1)*psz);
    munmap((void *)hib[devid].backend,     (hib[devid].backend_mem_bytes/psz+1)*psz);
    munmap((void *)hib[devid].piow_dblbuf, (hib[devid].piowbuf_bytes/psz+1)*psz);
    munmap((void *)hib[devid].dmar_buf,    ((hib[devid].dmabuf_bytes-1)/psz+1)*psz);
    munmap((void *)hib[devid].dmaw_buf,    ((hib[devid].dmabuf_bytes-1)/psz+1)*psz);
#endif

    close(hib_fd[devid]);
    hib_fd[devid] = -1;
        WARN(3, "hib[%d] closed.\n", devid);
}

void hib_reset_dma(int devid)
{
    int i;
    //    hib_mem_writeMC(devid, hib[devid].command, (1<<hib[devid].command_reset_backend_bit));
    // the above should work, but seemingly not?
    // JM 20100403
    //    fprintf(stderr, "dev[%d]reg 28 before dma reset: 0x%8x\n",devid,
    //    hib_mem_readMC(devid, 0x28));
    hib_mem_writeMC(devid, 0x20, 0x80000000);
    //    fprintf(stderr, "dev[%d]reg 28 after dma reset: 0x%8x\n", devid,
    //	    hib_mem_readMC(devid, 0x28));
}    
void hib_send_resetMC(int devid, int reset_code)
{
    // current reset bits (as of 2010/4/8)
    // 31: DMA and DMA FIFO
    // 30: backend pipeline
    // 29: serial transievers
    // 28: transivers related logic
    //    hib_mem_writeMC(devid, hib[devid].command, (1<<hib[devid].command_reset_backend_bit));
    // the above should work, but seemingly not?
    // JM 20100403
    fprintf(stderr,"hib_send_resetmc writing 0x%x\n", reset_code << 28);
    hib_mem_writeMC(devid, 0x20, reset_code << 28);
}    

void
hib_set_test_modeMC(int devid, int mode)
{
    hib_mem_writeMC(devid, hib[devid].testmode, mode);
}

void
hib_set_sendfuncMC(int devid, int func)
{
    switch (func) {
      case SENDFUNC_PIOW:
	  SendMCfunc[devid] = hib_piowMC;
	  break;
      case SENDFUNC_DMAR:
	  SendMCfunc[devid] = hib_dmarMC;
	  break;
      default:
	  break;
    }
}

void
hib_sendMC(int devid, int size, UINT64 *buf) /* size: transfer size in 8-byte word */
{
  SendMCfunc[devid](devid, size, buf);
}

#define PREFETCH(mem)     asm volatile("prefetchnta %0"        : : "m"(mem))
#define LOADDQ(mem, reg)  asm volatile("movdqa %0, %"reg       : : "m"(mem));
#define STOREDQ(reg, mem) asm volatile("movntdq %"reg " , %0"  : : "m"(mem));

static void
streamwrite16(int size, UINT64 *src, UINT64 *dst) // size in 8-byte word
{
    int j;

    // args need to be 16-byte aligned.
    assert(((unsigned long)src & 0xf) == 0);
    assert(((unsigned long)dst & 0xf) == 0);

    for (j = 0; j < size; j += 2) {
        PREFETCH(src[j+2]);
        LOADDQ(src[j], "%xmm0");
        STOREDQ("%xmm0", dst[j]);
    }
}

void
hib_piowMC(int devid, int size, UINT64 *buf) /* size: transfer size in 8-byte word */
{
    int i, j;
    int s = (hib[devid].piowbuf_bytes>>3);

    s = 128;
    //s = 256;
    //s = 512;

    for (i = 0; i < size; i += s) {

      if (i + s > size) {
	s = size - i;
      }

#if 1
      for (j = 0; j < s; j++) {
          hib[devid].piow_dblbuf[j] = buf[i + j];
      }
#else
      streamwrite16(s, buf + i, (UINT64 *)hib[devid].piow_dblbuf);
#endif


      if (hib[devid].type == HIB_HIBDRV) {
          hib_mem_writeMC(devid, hib[devid].command,
                          (1<<(hib[devid].command_swap_sram_bit)) |
                          ((s*8)<<(hib[devid].command_sram_wcnt_bit))); // s in byte unit.
          // asm volatile("mfence":::"memory");
      }
      else {
          hib_mem_writeMC(devid, hib[devid].dma0misc,
                          (1<<(hib[devid].dma0misc_swap_sram_bit)) |
                          (s<<(hib[devid].dma0misc_sram_wcnt_bit))); // s in 8-byte unit.
      }
    }
}


static inline UINT64 rdtsc2()
{
	UINT64 ret;
	asm __volatile__("rdtsc  \n\t" \
			 "shlq $32, %%rdx \n\t" \
			 "orq  %%rdx, %%rax \n\t" \
			 "movq %%rax, %0" : "=g"(ret) : :"rax", "rdx");
	return ret;
}

static void xwait(int wait)
{
    UINT64 st = rdtsc2();
    while((rdtsc2()-st)<wait) {}
}

void hib_xwait(int wait)
{
    xwait(wait);
}

/* DMA read (host -> hib) using DMA0 channel */
// NOTE added by JM 2010/1/9
// This fuction seems to be rotten. hib->dmar_buf seems to be never used...
// 
void hib_dmarMC(int devid, int size, UINT64 *buf)
/* size: transfer size in 8-byte word */
{
    volatile unsigned long misc;
    unsigned long offset;

    if (size > (hib[devid].dmabuf_bytes>>3)) {
	fprintf(stderr, "hib_dmarMC(): size too large.\n");
	fprintf(stderr, "size: %d\n", size);
	exit(2);
    }

    // !!! may not be necessary.
    if (size < 8) {
	int ii;
	for (ii = size; ii < 8; ii++) {
	    buf[ii] = 0; // fill buf with dummy data
	}
	size = 8;
    }

    offset = dmar_pa[devid];
    hib_mem_writeMC(devid, hib[devid].dma0lsb, offset);
    hib_mem_writeMC(devid, hib[devid].dma0size, size*8); /* size in byte */
    hib_mem_writeMC(devid, hib[devid].dma0cmd, 1 << hib[devid].dma0cmd_kickoff_bit);

    //    __mm_memfence();

    /*
     * wait till the DMA transfer completes
     * dmadone(1:0) is mapped to DMADONE(1:0)
     */
    //    fprintf(stderr,"DMA read wait start\n");
    do {
	xwait(5000);
	misc = hib_mem_readMC(devid, hib[devid].dma0misc);
    } while ((misc & (1<<hib[devid].dma0misc_done_bit)) == 0); /* wait until dmadone0 is asserted */
}

static int dmar_length[MAXHIBDEVS]={0,0,0,0,0,0,0,0};
static int dmar_length_prev[MAXHIBDEVS]={0,0,0,0,0,0,0,0};
static unsigned long dmar_offset_prev[MAXHIBDEVS]={0,0,0,0,0,0,0,0};
static int dmar_running[MAXHIBDEVS]={0,0,0,0,0,0,0,0};
static int dmar_current_plane[MAXHIBDEVS] ={0,0,0,0,0,0,0,0};
#define DMAR_MAXSIZE 10240


void hib_start_dmarMC(int devid, int size, UINT64 *buf)
/* size: transfer size in 8-byte word */
// Note by JM 2010/1/9: This function is broken. Should not be used
// argument buf is used nowhere.
{
    unsigned long offset;
    
    if (size > (hib[devid].dmabuf_bytes>>3)) {
	fprintf(stderr, "hib_dmarMC(): size too large.\n");
	fprintf(stderr, "size: %d\n", size);
	exit(2);
    }
    
    // !!! may not be necessary
    if (size < 8) {
	int ii;
	for (ii = size; ii < 8; ii++) {
	    buf[ii] = 0; // fill buf with dummy data
	}
	size = 8;
    }
    
    offset = dmar_pa[devid];
    hib_mem_writeMC(devid, hib[devid].dma0lsb, offset);
    hib_mem_writeMC(devid, hib[devid].dma0size, size*8); /* size in byte */
    hib_mem_writeMC(devid, hib[devid].dma0cmd, 1 << hib[devid].dma0cmd_kickoff_bit);
    //    __mm_memfence();
}

void hib_start2_dmarMC_multiplane_bcast(int devid, int bcastmode)
// bcastmode = 0: no bcast
//             n: (n>0) bcast to boards modulo n
//                1: same data sent to all board
//                2: 0,2,4... receive same data, and 1,3,5... receive same data

{
    static int last_bcastmode[MAXHIBDEVS] = {-1,-1,-1,-1,-1,-1,-1,-1};
    int bcastdevid = bcastmode? (devid%bcastmode):devid;

    //    fprintf(stderr,"start DMAR %d %d plane %d\n", devid, size, dmar_current_plane[devid]);
    if(dmar_running[devid]){
	//fprintf(stderr,"DMAR running on dev %d- wait to finish\n", devid);
	hib_finish_dmarMC(devid);
    }
    int size = dmar_length[bcastdevid];
    if (bcastdevid != devid){
	// this is in real bcast. Since the device bcastdevid
	// has DMAR already started, we need to pick the size from
	// its previous size
	size = dmar_length_prev[bcastdevid];
	if (dmar_current_plane[devid] == dmar_current_plane[bcastdevid]){
	    // this should not occur...
	    fprintf(stderr,"hib_start2_dmarMC_multiplane_bcast");
	    fprintf(stderr," dmar planes %d %d inconsistent\n",
		    dmar_current_plane[devid],dmar_current_plane[bcastdevid]);
	    fprintf(stderr," for devids %d %d\n",devid, bcastdevid);
	}
    }
		    
    if ((size != dmar_length_prev[devid]) ||(bcastmode != last_bcastmode[devid])){
	//	fprintf(stderr,"start(2) register new dma seq DMAR %d %d\n", devid, size);
	//	fprintf(stderr,"bcastmode=%d bcatid=%d\n", bcastmode, bcastdevid);
	hib_register_dma_seq_bcastMC(devid, bcastdevid, 0, 0, size,
				     hib[bcastdevid].dmar_buf);
	hib_register_dma_seq_bcastMC(devid, bcastdevid, 1, 0, size,
			       hib[bcastdevid].dmar_buf+DMAR_MAXSIZE);
    }
    hib_start_dma_seqMC(devid,dmar_current_plane[devid] ,0);
    
    dmar_running[devid]=1;
    //    fprintf(stderr,"old plane %d, %d\n", devid, dmar_current_plane[devid]);
    dmar_current_plane[devid]= 1-dmar_current_plane[devid];
    dmar_length[devid]=0;	
    dmar_length_prev[devid]=size;
    //    fprintf(stderr,"new plane %d, %d\n", devid, dmar_current_plane[devid]);
    last_bcastmode[devid] = bcastmode;
}

void hib_start2_dmarMC_multiplane(int devid)
{
    hib_start2_dmarMC_multiplane_bcast(devid,0);
}

void hib_start2_dmarMC_old(int devid)

{
    unsigned long offset;
    int size = dmar_length[devid];
    offset = dmar_pa[devid];
    //    fprintf(stderr,"start DMAR %d %d\n", devid, size);
    hib_mem_writeMC(devid, hib[devid].dma0lsb, offset);
    hib_mem_writeMC(devid, hib[devid].dma0size, size*8); /* size in byte */
    hib_mem_writeMC(devid, hib[devid].dma0cmd, 1 << hib[devid].dma0cmd_kickoff_bit);
    dmar_running[devid]=1;
    dmar_length_prev[devid]=size;
    dmar_offset_prev[devid]=offset;
    dmar_length[devid]=0;	
}

void hib_start2_dmarMC(int devid)

{
    unsigned long offset;
    int size = dmar_length[devid];
    offset = dmar_pa[devid]; 
    //    fprintf(stderr,"start(2) DMAR %d %d\n", devid, size);
    if ((size != dmar_length_prev[devid])||(offset !=  dmar_offset_prev[devid])){
	// need to register DMAR table
	fprintf(stderr,"start(2) register new dma seq DMAR %d %d\n", devid, size);
	hib_register_dma_seqMC(devid, 0, 0, size, hib[devid].dmar_buf);
    }
    if (0){
	hib_mem_writeMC(devid, hib[devid].dma0lsb, offset);
	hib_mem_writeMC(devid, hib[devid].dma0size, size*8); /* size in byte */
	hib_mem_writeMC(devid, hib[devid].dma0cmd, 1 << hib[devid].dma0cmd_kickoff_bit);
    }
    hib_start_dma_seqMC(devid, 0,0);
    dmar_running[devid]=1;
    dmar_length_prev[devid]=size;
    dmar_offset_prev[devid]=offset;
    dmar_length[devid]=0;	
}

int hib_register_data_dmarMC_multiplane(int devid, int size, UINT64 *buf)
{
    //    fprintf(stderr, "hib_register_data_dmarMC called for %d %d\n",
    //	    devid, size);
    //    fprintf(stderr,"current buf, plane = %d %d\n",
    //	    dmar_length[devid], dmar_current_plane[devid]);
    //    if (size+dmar_length[devid] > DMAR_MAXSIZE ){
    //	hib_start2_dmarMC(devid);
    //	hib_clear_dmarMC(devid);
    //    }
    int i;
    UINT64 *dest = hib[devid].dmar_buf + dmar_current_plane[devid]*DMAR_MAXSIZE
	+dmar_length[devid];
    
    UINT64 *src = buf;
    //    for(i=0;i<size;i++)dest[i]=src[i];
    hibutil_memcopy64(src, dest, size);
    dmar_length[devid]+= size;
    return 0;
}

int hib_register_data_dmarMC(int devid, int size, UINT64 *buf)
{
    //    fprintf(stderr, "hib_register_data_dmarMC called for %d %d\n",
    //	    devid, size);
    //    fprintf(stderr,"current buf, plane = %d %d\n",
    //	    dmar_length[devid], dmar_current_plane[devid]);
    int i;
    UINT64 *dest = hib[devid].dmar_buf 	+dmar_length[devid];
    UINT64 *src = buf;
    //    for(i=0;i<size;i++)dest[i]=src[i];
    hibutil_memcopy64(src, dest, size);
    dmar_length[devid]+= size;
    return 0;
}


int hib_status_dmarMC(int devid)
{
    return dmar_running[devid];
}
int hib_bufferd_words_dmarMC(int devid)
{
    return dmar_length[devid];
}

void hib_finish_dmarMC_multiplane(int devid)
{
    volatile unsigned long misc;
    /*
     * wait till the DMA transfer completes
     * dmadone(1:0) is mapped to DMADONE(1:0)
     */
    do {
	misc = hib_mem_readMC(devid, hib[devid].dma0misc);
	//	fprintf (stderr,"hib_finish_dmarMC(%d) = %x\n", devid, misc);
    } while ((misc & (1<<hib[devid].dma0misc_done_bit)) == 0);
    /* wait until dmadone0 is asserted */
    dmar_running[devid]=0;
}

static int dmar_timeout = 10000000;

void hib_finish_dmarMC(int devid)
{
    volatile unsigned long misc;
    /*
     * wait till the DMA transfer completes
     * dmadone(1:0) is mapped to DMADONE(1:0)
     */
    int i = 0;

    //    fprintf(stderr,"hib_finish_dmarMC(%d)\n", devid);
    //    printf("hib_finish_dmarMC\n");
    
    do {
	xwait(500);
	misc = hib_mem_readMC(devid, hib[devid].dma0misc);
	//	fprintf (stderr,"hib_finish_dmarMC(%d) = %x\n", devid, misc);
	i += 50;
	if (i > dmar_timeout){
	    fprintf(stderr,"dmar timeout on devid=%d\n", devid);
	    exit(-1);
	}
    } while ((misc & (1<<hib[devid].dma0misc_done_bit)) == 0);
    /* wait until dmadone0 is asserted */
    dmar_running[devid]=0;
}


void hib_clear_dmarMC(int devid)
{
    dmar_length[devid]=0;
}

void hib_clear_dmarMC_multiplane(int devid)
{
    //    fprintf(stderr,"clea_dmarMC called for %d\n", devid);
    if (dmar_running[devid]) hib_finish_dmarMC(devid);
    dmar_length[devid]=0;
    dmar_current_plane[devid]=0;
}



/* DMA write (host <- hib) using DMA1 channel */
int
hib_dmawMC(int devid, int size, UINT64 *buf) // size in 8-byte word
{
    hib_start_dmawMC(devid, size, buf);
    return hib_finish_dmawMC(devid);
}

/* start DMA write (host <- hib) using DMA1 channel */
#define NAN_AT_EOD (0x12345678ULL)
UINT64 *eobuf[NHIB];
UINT64 *seqeobuf[NDMAWSEQ][NHIB];
UINT64 *seqhobuf[NDMAWSEQ][NHIB];
UINT64 *startbuf[NHIB];
int dmasize[NHIB];

static int dmacheck_mode = 1;
static int dmacheck_stride = 32;
int hib_set_dmacheck_mode(int mode)
{
    dmacheck_mode =mode;
}

int hib_get_dmacheck_mode()
{
    return dmacheck_mode;
}

void hib_initialize_dmawbufferMC(int devid, int size, UINT64 *buf) 
{
    int i;
    for(i=0;i<size;i+=dmacheck_stride) buf[i] = NAN_AT_EOD;
}

void
hib_start_dmawMC(int devid, int size, UINT64 *buf) // size in 8-byte word
{
    unsigned long offset;
    static int firstcall[8] = {
        1, 1, 1, 1, 1, 1, 1, 1, 
    };
    if (size > (hib[devid].dmabuf_bytes>>3)) {
        fprintf(stderr, "hib_dmawMC(): size too large.\n");
        fprintf(stderr, "size: %d\n", size);
        exit(2);
    }

    offset = buf - hib[devid].dmaw_buf; // offset in 8-byte word.

    if (offset < 0 || hib[devid].dmabuf_bytes < offset * 8) {
	fprintf(stderr, "hib_start_dmawMC: address offset (0x%016llx) out of range.\n", offset);
	exit(2);
    }
    buf[0]=-1;
    // set a dummy data at the end of buffer.
    buf[size - 1] = NAN_AT_EOD;
    buf[size/2 - 1] = NAN_AT_EOD;
    //  this part is now should be done by application side, by
    // calling hib_initialize_dmawbufferMC explicitly before
    // calling hib_start_dmawMC
    //    if (dmacheck_mode){
    //	// fill NAN_AT_EOD for every one in dmacheck_stride words
    //	int i;
    //	for(i=0;i<size;i+=dmacheck_stride) buf[i] = NAN_AT_EOD;
    //	//	for(i=0;i<size;i+=32) __builtin_ia32_movntq(buf+i,NAN_AT_EOD);
    //    }
    eobuf[devid] = &(buf[size - 1]);
    startbuf[devid] = buf;
    dmasize[devid]=size;

    if (hib[devid].type == HIB_HIBDRV) {
#if 1
        // write two 32-bit registers by a single 64-bit write access.
        // write PCI lower 32-bit address register and transfer size register at once,
        // and then kick off a DMA transfer.
        hib_mem_write64MC(devid, hib[devid].dma1lsb,
                          ((UINT64)(size*8) << 32) | (dmaw_pa[devid] + offset * 8));
        //        asm volatile("mfence":::"memory");
#else

        // hib_mem_writeMC(devid, hib[devid].dma1msb, 0x0);
        //        hib_mem_writeMC(devid, hib[devid].dma1laddr, 0x0); /* start address in local space */
        //        hib_mem_writeMC(devid, hib[devid].dma1cmd, 0); // set default param

        // lower 32-bit of start address in PCI space.
        hib_mem_writeMC(devid, hib[devid].dma1lsb, dmaw_pa[devid] + offset * 8);
        // size in byte and then start DMA.
        hib_mem_writeMC(devid, hib[devid].dma1size, size*8);
#endif
    }
    else {
        hib_mem_writeMC(devid, hib[devid].dma1lsb, dmaw_pa[devid] + offset * 8); /* lower 32-bit of start address in PCI space */
        hib_mem_writeMC(devid, hib[devid].dma1msb, 0x0); /* higher 32-bit of start address in PCI space */
        hib_mem_writeMC(devid, hib[devid].dma1size, size*8); /* size in byte */
        hib_mem_writeMC(devid, hib[devid].dma1cmd, 1<<hib[devid].dma1cmd_kickoff_bit);
//    fprintf(stderr,"dmaw started\n");
//        hib_dump_sreg(devid);
    }
}

void hib_set_dma_extended_regMC(int devid, int rammode, int bufsize,
				int dmacmask)
{
    int data = (bufsize & 0xff) | ((rammode &1)<<8)|((dmacmask &0x1f)<<9);
    hib_mem_writeMC(devid, HIB_DMA_ADRREG, 0x40);
    hib_mem_writeMC(devid, HIB_DMA_DATAREG, data);
}


void hib_start_dma_seqMC(int devid, int seq_id, int wmode)
// wmode=0: DMA Read else write
{
    int trigger;
    if (wmode){
	hib_setup_dma_eodMC(devid, seq_id);
	trigger = 1 << (seq_id + 4);
    }else{
	trigger = 1 << (seq_id);
    }
    hib_mem_writeMC(devid, HIB_DMA_TRIGREG, trigger);
}
void hib_setup_dma_eodMC(int devid, int seq_id)
{
    UINT64 * buf = dmaw_seq[devid][seq_id].buf;
    int size = dmaw_seq[devid][seq_id].size;
    buf[0] = -1;

    buf[size - 1] = NAN_AT_EOD;
    buf[size/2 - 1] = NAN_AT_EOD;
    eobuf[devid] = &(buf[size - 1]);
    seqeobuf[seq_id][devid] = &(buf[size - 1]);
    seqhobuf[seq_id][devid] = &(buf[size/2 - 1]);
    startbuf[devid] = buf;
    dmasize[devid]=size;
}

void hib_register_dma_seq_bcastMC(int devid, int bcastdevid, int seq_id,
				  int wmode, int size, UINT64 *buf)
// wmode=0: DMA Read else write
{
    unsigned long offset;
    if (size > (hib[devid].dmabuf_bytes>>3)) {
        fprintf(stderr, "hib_register_dma_seqMC(): size too large.\n");
        fprintf(stderr, "size: %d\n", size);
        exit(2);
    }
    UINT64 *buf_base;
    unsigned long pa_base;
    int start_adr_base;
    int dma_param;

    //    fprintf(stderr,"hib_register_dma_seqMC devide=%d seq=%d, mode, size=%d %d\n",
    //	    devid, seq_id, wmode, size);

    if (wmode){
	buf_base = hib[devid].dmaw_buf;
	pa_base = dmaw_pa[devid];
	start_adr_base = 0x20;
	dmaw_seq[devid][seq_id].buf = buf;
	dmaw_seq[devid][seq_id].size = size;
	dma_param=0x0700;
    }else{
	buf_base = hib[bcastdevid].dmar_buf;
	pa_base = dmar_pa[bcastdevid];
	start_adr_base = 0x0;
	dmar_seq[devid][seq_id].buf = buf;
	dmar_seq[devid][seq_id].size = size;
	//	dma_param=0x0600; // still in the fifo mode
	dma_param=0x0601; // test with ram mode
    }	
    
    offset = buf - buf_base; // offset in 8-byte word.

    if (offset < 0 || hib[devid].dmabuf_bytes < offset*8 ) {
	fprintf(stderr,"hib_register_dma_seqMC: address offset (0x%016llx) out of range.\n",
		offset);
	exit(2);
    }
	
    int start_address = start_adr_base + 0x08 * seq_id;
    //    fprintf(stderr, "hib_register_dma_seqMC\n");
    //    fprintf(stderr, "start_address = 0x%lx\n", start_address);
    //    fprintf(stderr, "dma address low = %lx: 0x%lx\n", start_address,
    //	    pa_base + offset * 8);
    //    fprintf(stderr, "dma size= %lx: 0x%lx\n", start_address+2, size*8);
    hib_mem_writeMC(devid, HIB_DMA_ADRREG, start_address);
    hib_mem_writeMC(devid, HIB_DMA_DATAREG, pa_base + offset * 8);
    hib_mem_writeMC(devid, HIB_DMA_ADRREG, start_address+1);
    hib_mem_writeMC(devid, HIB_DMA_DATAREG, 0x0);
    hib_mem_writeMC(devid, HIB_DMA_ADRREG, start_address+2);
    hib_mem_writeMC(devid, HIB_DMA_DATAREG, size*8);
    hib_mem_writeMC(devid, HIB_DMA_ADRREG, start_address+3);
    hib_mem_writeMC(devid, HIB_DMA_DATAREG, 0x0);
    hib_mem_writeMC(devid, HIB_DMA_ADRREG, start_address+4);
    hib_mem_writeMC(devid, HIB_DMA_DATAREG, dma_param);
}
void hib_register_dma_seqMC(int devid, int seq_id, int wmode,
			    int size, UINT64 *buf)
// wmode=0: DMA Read else write
{
    unsigned long offset;
    if (size > (hib[devid].dmabuf_bytes>>3)) {
        fprintf(stderr, "hib_register_dma_seqMC(): size too large.\n");
        fprintf(stderr, "size: %d\n", size);
        exit(2);
    }
    UINT64 *buf_base;
    unsigned long pa_base;
    int start_adr_base;
    int dma_param;

    //    fprintf(stderr,"hib_register_dma_seqMC devide=%d seq=%d, mode, size=%d %d\n",
    //	    devid, seq_id, wmode, size);

    if (wmode){
	buf_base = hib[devid].dmaw_buf;
	pa_base = dmaw_pa[devid];
	start_adr_base = 0x20;
	dmaw_seq[devid][seq_id].buf = buf;
	dmaw_seq[devid][seq_id].size = size;
	dma_param=0x0700;
    }else{
	buf_base = hib[devid].dmar_buf;
	pa_base = dmar_pa[devid];
	start_adr_base = 0x0;
	dmar_seq[devid][seq_id].buf = buf;
	dmar_seq[devid][seq_id].size = size;
	//	dma_param=0x0600; // still in the fifo mode
	dma_param=0x0601; // test with ram mode
    }	
    
    offset = buf - buf_base; // offset in 8-byte word.

    if (offset < 0 || hib[devid].dmabuf_bytes < offset*8 ) {
	fprintf(stderr,"hib_register_dma_seqMC: address offset (0x%016llx) out of range.\n",
		offset);
	exit(2);
    }
	
    int start_address = start_adr_base + 0x08 * seq_id;
    //    fprintf(stderr, "hib_register_dma_seqMC\n");
    //    fprintf(stderr, "start_address = 0x%lx\n", start_address);
    //    fprintf(stderr, "dma address low = %lx: 0x%lx\n", start_address,
    //	    pa_base + offset * 8);
    //    fprintf(stderr, "dma size= %lx: 0x%lx\n", start_address+2, size*8);
    hib_mem_writeMC(devid, HIB_DMA_ADRREG, start_address);
    hib_mem_writeMC(devid, HIB_DMA_DATAREG, pa_base + offset * 8);
    hib_mem_writeMC(devid, HIB_DMA_ADRREG, start_address+1);
    hib_mem_writeMC(devid, HIB_DMA_DATAREG, 0x0);
    hib_mem_writeMC(devid, HIB_DMA_ADRREG, start_address+2);
    hib_mem_writeMC(devid, HIB_DMA_DATAREG, size*8);
    hib_mem_writeMC(devid, HIB_DMA_ADRREG, start_address+3);
    hib_mem_writeMC(devid, HIB_DMA_DATAREG, 0x0);
    hib_mem_writeMC(devid, HIB_DMA_ADRREG, start_address+4);
    hib_mem_writeMC(devid, HIB_DMA_DATAREG, dma_param);
}

/*
 * wait completion of the DMA write (host <- hib) transaction
 * kicked off by hib_start_dmawMC()
 */

static int dmaw_timeout = 10000000;

static int dmaw_timeour_error_occured = 0;
int hib_finish_dmawMC(int devid)
{
    volatile unsigned long misc;
    int ret = 0;
    dmaw_timeour_error_occured = 0;
    /* 
     * wait till the DMA transfer completes
     */

    // wait until a dummy data at the end of buffer is overwritten
    // by a data received from hib.
    int i;
    //    hib_dump_sreg(devid);

    for(i=0; i < dmaw_timeout; i++){
	if (*eobuf[devid] != NAN_AT_EOD) return 0;
	xwait(1000);
    }
    dmaw_timeour_error_occured = 1;
    fprintf(stderr,"DMAW timeout error for hib %d\n", devid);
    hib_dump_sreg(devid);
    ret = -1;
}

int hib_dmaw_timeout_error()
{
    return dmaw_timeour_error_occured;
}

int hib_finish_dmaw_seqMC(int devid, int seqid)
{
    // wait until a dummy data at the end of buffer is overwritten
    // by a data received from hib.
    int i;
    dmaw_timeour_error_occured = 0;
    for(i=0; i < dmaw_timeout; i++){
	if (*seqeobuf[seqid][devid]!= NAN_AT_EOD) return 0;
	xwait(1000);
    }
    dmaw_timeour_error_occured = 1;
    fprintf(stderr,"DMAW timeout error for hib %d\n", devid);
    return -1;
}

int hib_half_finish_dmaw_seqMC(int devid, int seqid)
{
    // wait until a dummy data at the end of buffer is overwritten
    // by a data received from hib.
    int i;
    dmaw_timeour_error_occured = 0;
    for(i=0; i < dmaw_timeout; i++){
	if (*seqhobuf[seqid][devid]!= NAN_AT_EOD) return 0;
	xwait(1000);
    }
    dmaw_timeour_error_occured = 1;
    fprintf(stderr,"DMAW timeout error for hib %d\n", devid);
    return -1;
}


int hib_remaining_words_dmawMC_simple(int devid)
{
    if (*eobuf[devid] != NAN_AT_EOD) return 0;
    if (dmacheck_mode == 0) return 1;
    int starti = ((dmasize[devid]-1)/dmacheck_stride)*dmacheck_stride;
    //    fprintf(stderr,"devid=%d, size, start = %d %d\n",
    //            devid, dmasize[devid], starti);
    int i;
    for(i=starti; i>= 0; i-= dmacheck_stride){
	if (startbuf[devid][i] != NAN_AT_EOD){
	    return dmasize[devid]-i-1;
	}
    }
    return dmasize[devid];
}


int hib_remaining_words_dmawMC_binarysearch(int devid)
{
    if (*eobuf[devid] != NAN_AT_EOD) return 0;
    if (dmacheck_mode == 0) return 1;
    int iend = (dmasize[devid]-1)/dmacheck_stride;
    int ifirst = 0;
    while ((iend - ifirst)>1){
	int i = (iend + ifirst)/2;
	if (startbuf[devid][i*dmacheck_stride] != NAN_AT_EOD){
	    ifirst = i;
	}else{
	    iend = i;
	}
	//	fprintf(stderr,"iend, ifirst, i=%d %d %d\n", iend, ifirst, i);
    }
    return dmasize[devid]-iend*dmacheck_stride;
}

int hib_remaining_words_dmawMC(int devid)
{
    return hib_remaining_words_dmawMC_binarysearch(devid);
}




/*
 * check if dmar is done or not yet.
 */
int
hib_dmar_is_completedMC(int devid)
{
    volatile unsigned long misc;
    int ret = 0; /* -1 on error */

    misc = hib_mem_readMC(devid, hib[devid].dma1misc);
    if ((misc & (1<<hib[devid].dma0misc_done_bit)) == 0) {
        ret = 0;
    }
    else {
        ret = 1;
    }
    return ret;
}


/*
 * check if dmaw is done or not yet.
 */
int
hib_dmaw_is_completedMC(int devid)
{
    volatile unsigned long misc;
    int ret = 0; /* -1 on error */

    misc = hib_mem_readMC(devid, hib[devid].dma1misc);
    if ((misc & (1<<hib[devid].dma1misc_done_bit)) == 0) {
        ret = 0;
    }
    else {
        ret = 1;
    }
    return ret;
}

UINT32
hib_config_readMC(int devid, UINT32 addr)
{
    struct long_access ca;
    ca.addr = addr;
    ca.data = 0;

    if (hib_fd[devid] < 0) {
	fprintf(stderr, "open hib[%d] first\n", devid);
	exit (1);
    }
    ioctl(hib_fd[devid], hib[devid].read_cfg, &ca);
    return (ca.data);
}


void
hib_config_writeMC(int devid, UINT32 addr, UINT32 value)
{
    struct long_access ca;
    ca.addr = addr;
    ca.data = value;

    if (hib_fd[devid] < 0) {
	fprintf(stderr, "open hib[%d] first\n", devid);
	exit (1);
    }
    ioctl(hib_fd[devid], hib[devid].write_cfg, &ca);
}


// set some reasonable default for maxrequest
// (current value is 2, 1KB max?)
static void hib_config_setmaxrequestMC0(int devid)
{
    UINT32 val = hib_config_readMC(devid, 0x88);
    val |= 0x7000; // max possible value
    val &= 0xffff0fff;
    val |= 0x4000; //changed from 2 to 3 on 2010/5/19
                   // and to 4 on 2010/10/3
    //       fprintf(stderr,"set 0x88 to 0x%x\n", val);
    hib_config_writeMC(devid, 0x88, val);
}
void hib_config_setmaxrequestMC(int devid, int size)
{
    UINT32 val = hib_config_readMC(devid, 0x88);
    val |= 0x7000; // max possible value
    val &= 0xffff0fff;
    val |= ((size & 0x7)<<12);
       fprintf(stderr,"set 0x88 to 0x%x\n", val);
    hib_config_writeMC(devid, 0x88, val);
}


    


void hib_mem_writeMC(int devid, UINT32 addr, UINT32 value)
{
    WARN(4, "hib[%d] hib_mem_writeMC: 0x%08x pa: %08x\n",
         devid, addr, value);

    //    asm volatile("mfence":::"memory");
    hib[devid].local_reg[addr>>2] = value; // bar0, 32-bit address
    //    asm volatile("mfence":::"memory");

    //    hib[devid].backend[addr>>3] = value; // bar1, 64-bit address
    //    hib[devid].piow_dblbuf[addr>>3] = value; // bar2, 64-bit address
}

UINT32
hib_mem_readMC(int devid, UINT32 addr)
{
    UINT32 ret;

    ret =  hib[devid].local_reg[addr>>2]; // bar0, 32-bit address
    // ret =  hib[devid].backend[addr>>3]; // bar1, 64-bit address
    // ret =  hib[devid].piow_dblbuf[addr>>3]; // bar2, 64-bit address
    return (ret);
}

void
hib_mem_write64MC(int devid, UINT32 addr, UINT64 value)
{
    //    asm volatile("mfence":::"memory");
    *((UINT64 *)&(hib[devid].local_reg[addr>>2])) = value;
    //    asm volatile("mfence":::"memory");
}

UINT64
hib_mem_read64MC(int devid, UINT32 addr)
{
    UINT64 ret;

    ret =  *((UINT64*)&(hib[devid].local_reg[addr>>2]));

    return (ret);
}

static int dummy_val;

void hib_flush_piow_channel(int devid)
// expect pio channel to be flushed
// by issuing a read access
//
{
    //    dummy_val = hib_mem_readMC(devid, 0x18);
    hib_mem_writeMC(devid, 0x60, 0x00);
}


void hib_dump_sreg(int devid)
{
    int i,j;
    fprintf(stderr,"hib %d hibreg 0x18 = %8x %8x dma1 regs:\n", devid,
	    hib_mem_readMC(devid, 0x1c), hib_mem_readMC(devid, 0x18));
    fprintf(stderr,"hib %d hibreg 0x28 = %8x %8x misc regs:\n", devid,
	    hib_mem_readMC(devid, 0x2c), hib_mem_readMC(devid, 0x28));
    for(i=0;i<16;i++){
	hib_mem_writeMC(devid, 0x0, i);
	fprintf(stderr," [%1x]: 0x%08x ", i, hib_mem_readMC(devid, 0x60));
	if ((i%4)==3)fprintf(stderr,"\n");
    }
}


void TBgetDmaInfo(int devid,
		  unsigned long *dmarpa, UINT64 **dmarva,
		  unsigned long *dmawpa, UINT64 **dmawva)
/*
 * pa: physical addr of DMA read/write buf alloced in the kernel space
 * va: virtual addr (in the user process context) pa is mapped to
 */
{
    size_t psz = getpagesize();
    
    /* map DMA read buffer */
    if (-1 == ioctl(hib_fd[devid], hib[devid].get_dmar_pa, dmarpa)) {
	perror("TBgetDmaInfo() failed");
	exit (1);
    };
    if (-1 == ioctl(hib_fd[devid], hib[devid].set_map_mode, hib[devid].map_dmarbuf)) {
	perror("TBgetDmaInfo() failed");
	exit (1);
    };
    *dmarva = mmap(NULL,
		   ((hib[devid].dmabuf_bytes-1)/psz+1)*psz,
		   PROT_READ | PROT_WRITE,
		   MAP_SHARED,
		   hib_fd[devid],
		   0);
    if ((long int)(*dmarva) == -1) {
	perror("TBgetDmaInfo");
	exit (1);
    }
    
    /* map DMA write buffer */
    if (-1 == ioctl(hib_fd[devid], hib[devid].get_dmaw_pa, dmawpa)) {
	perror("TBgetDmaInfo() failed");
	exit (1);
    };
    if (-1 == ioctl(hib_fd[devid], hib[devid].set_map_mode, hib[devid].map_dmawbuf)) {
	perror("TBgetDmaInfo() failed");
	exit (1);
    };
    *dmawva = mmap(NULL,
		   ((hib[devid].dmabuf_bytes-1)/psz+1)*psz,
		   PROT_READ | PROT_WRITE,
		   MAP_SHARED,
		   hib_fd[devid],
		   0);
    if ((long int)(*dmawva) == -1) {
	perror("TBgetDmaInfo");
	exit (1);
    }
    if (-1 == ioctl(hib_fd[devid], hib[devid].set_map_mode, hib[devid].map_hibmem)) {
	perror("TBgetDmaInfo() failed");
	exit (1);
    };

    // shift read addresses by one word, to make the body of message aligned to 16-byte
    // boundary -- JM 2010/5/20
    *dmarpa++;
    *dmarva++;
}

static void TBmemMap(int devid)
{
    size_t psz = getpagesize();

    /*
     * maps HIB local register (BAR0)
     */
    if (-1 == ioctl(hib_fd[devid], hib[devid].set_map_mode, hib[devid].map_hibmem)) {
	perror("TBgetDmaInfo() failed");
	exit (1);
    };
    hib[devid].local_reg = mmap(NULL,
                                 (hib[devid].local_mem_bytes/psz+1)*psz,
                                 PROT_READ | PROT_WRITE,
                                 MAP_SHARED,
                                 hib_fd[devid],
                                 0); /* device driver takes care of bar offset */
    if ((long int)(hib[devid].local_reg) == -1) {
	fprintf(stderr, "%s%d\n", hib[devid].devname, devid);
	exit (1);
    }
    WARN(3, "hib[%d].local_reg: 0x%016lx\n", devid, hib[devid].local_reg);


    /*
     * maps BAR1 (HIB backend)
     */
    if (-1 == ioctl(hib_fd[devid], hib[devid].set_map_mode, hib[devid].map_backend)) {
	perror("TBmemMap() failed");
	exit (1);
    };
    hib[devid].backend = mmap(NULL,
                              (hib[devid].backend_mem_bytes/psz+1)*psz,
                              PROT_READ | PROT_WRITE,
                              MAP_SHARED,
                              hib_fd[devid],
                              0); /* device driver takes care of bar offset */
    if ((long int)(hib[devid].backend) == -1) {
	fprintf(stderr, "%s%d\n", hib[devid].devname, devid);
	exit (1);
    }
    WARN(3, "hib[%d].backend: 0x%016lx\n", devid, hib[devid].backend);


    /*
     * maps BAR2 (double buffer for PIO write transfer)
     */
    if (-1 == ioctl(hib_fd[devid], hib[devid].set_map_mode, hib[devid].map_dblbuf)) {
	perror("TBmemMap() failed");
	exit (1);
    };
    hib[devid].piow_dblbuf = mmap(NULL,
                                   (hib[devid].piowbuf_bytes/psz+1)*psz,
                                   PROT_READ | PROT_WRITE,
                                   MAP_SHARED,
                                   hib_fd[devid],
                                   0); /* device driver takes care of bar offset */
    if ((long int)(hib[devid].backend) == -1) {
	fprintf(stderr, "%s%d\n", hib[devid].devname, devid);
	exit (1);
    }
    WARN(3, "hib[%d].piow_dblbuf: 0x%016lx\n", devid, hib[devid].piow_dblbuf);


    /* house keeping */
    if (-1 == ioctl(hib_fd[devid], hib[devid].set_map_mode, hib[devid].map_hibmem)) {
	perror("TBgetDmaInfo() failed");
	exit (1);
    };
}

static int devid_offset = 0;
void hib_set_devid_offset(int offset){devid_offset = offset;}

static void
init_envs(int devid)
{
    char *p;
    int i, j;
#if NHIB > 8
#error NHIB must not exceed 8
#endif
    static int firstcall[8] = {
        1, 1, 1, 1, 1, 1, 1, 1, 
    };

    if (firstcall[devid]) {
	firstcall[devid] = 0;
        hib_fd[devid] = -1;

	for (j = 0, i = 0; j < NGRAPE7X; j++, i++) {
            if (i == devid) {
                sprintf(hib[i].devname, "/dev/%s%d", GRAPE7X_DEVNAME, j);
                hib[i].type = HIB_GRAPE7X;
            }
	}
	for (j = 0; j < NGRAPE7E; j++, i++) {
            if (i == devid) {
                sprintf(hib[i].devname, "/dev/%s%d", GRAPE7E_DEVNAME, j);
                hib[i].type = HIB_GRAPE7E;
            }
	}
	for (j = 0; j < NHIBDRV; j++, i++) {
            if (i == devid) {
                sprintf(hib[i].devname, "/dev/%s%d", HIBDRV_DEVNAME, j);
                hib[i].type = HIB_HIBDRV;
            }
	}
	p = getenv("HIB_WARNLEVEL");
        if (p) {
            int tmp, lv;
            tmp = atoi(strtok(p, " "));
            if (0 <= tmp) {
                lv = tmp;
                hib_set_warn_level(lv);
            }
        }
    }
}

static void
lock_hib(int devid)
{
    int locked; // 0:success  -1:failure
    int cnt = 0;

    locked = lockf(hib_fd[devid], F_TLOCK, 0);
    if (locked == 0) return; // returns successfully.

    if ((errno == EACCES) || (errno == EAGAIN)) {
	WARN(1, "Someone is using hib[%d]. Sleep...\n", devid);
    }
    else {
	perror("hib_openMC");
	exit(2);
    }
    locked = lockf(hib_fd[devid], F_LOCK, 0);
}

static int
try_lock_hib(int devid)
{
    int locked; // 0:success  -1:failure
    int cnt = 0;

    locked = lockf(hib_fd[devid], F_TLOCK, 0);
    if (locked == 0) return 0; // returns successfully.

    if ((errno == EACCES) || (errno == EAGAIN)) {
	WARN(3, "Someone is using hib[%d]. Sleep...\n", devid);
        return -1;
    }
    else {
	perror("hib_openMC");
	exit(2);
    }
}

static void
unlock_hib(int devid)
{
    if (lockf(hib_fd[devid], F_ULOCK, 0)) {
	perror("unlock_hibMC");
	exit (2);
    }
}
