#ifndef  PIT_PARTICLE_H
#  define   PIT_PARTICLE_H
/*-----------------------------------------------------------------------------
 *  pit_particle : basic class for simple nbody1-like program
 *  J. Makino 2001/3/11
 *-----------------------------------------------------------------------------
 */
#include "vector.h"

#define THREED
#define REAL_GRAVITY


class particle
    {
    private:
        vector pos;
	vector vel;
	vector acc;
	real   pot;
	real   mass;
	real   time;           //
	real   timestep;       //
	real   pot;            // potential
	vector jerk;           // (d/dt) acc
	vector old_acc;        //
	vector old_jerk;       //
	vector k_over_18;	// for 5th-order prediction (GRAPE-6)
	vector pred_pos;       // current predicted pos
	vector pred_vel;       // current predicted vel
	int    index;
    public:
	particle(){
	    pos = 0.0;
	    vel = 0.0;
	    acc = 0.0;
	    pot =  mass =  0.0;
	    index = 0;
	    time = timestep = pot = 0;
	    k_over_18 = jerk = old_acc = old_jerk = pred_pos = pred_vel = 0;
	    t_pred = -VERY_LARGE_NUMBER;
	}
	void set_timestep(real dt) {		// dt is always timestep
	    timestep = dt;
	}

	inline void set_time(real t) {		// t is always time
	    time = t;
	}

	// The "get_time" accessors should be OK given the above "set"
	// functions.

	inline real get_timestep()	  {return timestep;}
	inline real get_time()           {return time;}

	void  clear_interaction()         {acc = jerk = 0;
					   pot = 0;}

	void  set_acc_and_jerk_and_pot(const vector& a,
				       const vector& j, real p)
	   {acc = a; jerk = j; pot = p;}

	inline vector  get_pred_pos()     {predict_loworder(get_system_time());
					   return pred_pos;}
	inline vector  get_pred_vel()     {predict_loworder(get_system_time());
					   return pred_vel;}

	void  clear_t_pred()		  {t_pred = -VERY_LARGE_NUMBER;}
	inline xreal  get_t_pred()        {return t_pred;}
	
	real  get_pot()                   {return pot;}
	void  set_pot(const real p)	  {pot = p;}
	void  clear_pot()                 {pot = 0;}
	void  inc_pot(const real& d_pot)  {pot += d_pot;}

        inline void set_jerk(const vector& new_jerk)     {jerk = new_jerk;}
        inline void set_old_acc(const vector& new_acc)   {old_acc = new_acc;}
        inline void set_old_jerk(const vector& new_jerk) {old_jerk = new_jerk;}

	inline void clear_jerk()                         {jerk = 0;}

	inline void scale_jerk(const real scale_factor)
	    {jerk *= scale_factor;}
	inline void scale_old_acc(const real scale_factor)
	    {old_acc *= scale_factor;}
	inline void scale_old_jerk(const real scale_factor)
	    {old_jerk *= scale_factor;}

	inline void inc_jerk(const vector& d_jerk)       {jerk += d_jerk; }
	inline void inc_old_acc(const vector& d_acc)     {old_acc += d_acc; }
	inline void inc_old_jerk(const vector& d_jerk)   {old_jerk += d_jerk; }

	inline vector get_jerk()                {return jerk;}
	inline vector get_old_acc()             {return old_acc;}
	inline vector get_old_jerk()            {return old_jerk;}
	inline vector get_k_over_18()		{return k_over_18;}

	void  init_pred() {
	    pred_pos = pos;
	    pred_vel = vel;
	    t_pred = get_time();
	}

	void  store_old_force() {
	    old_acc = acc;
	    old_jerk = jerk;
	}

	inline void predict_loworder(xreal t) {		// t is always time
	       if (t_pred < t) {
		    real dt;
		    dt = t - get_time();
		    real dt3 = dt * ONE_THIRD;
		    real dt2 = dt * 0.5;
		    pred_pos = ((old_jerk * dt3
				 + old_acc) * dt2 + vel) * dt + pos;
		    pred_vel = (old_jerk * dt2
				+ old_acc) * dt + vel;
		    
		    t_pred = t;
	       }
	}

	inline void predict_loworder5(xreal t) {	// t is always time

	    // 5th-order prediction of a top-level node (not checked)

	    if (t_pred < t) {
		real dt = t - get_time();
		real dt3 = dt * ONE_THIRD;
		real dt2 = dt * 0.5;
		pred_pos = (((4.5 * k_over_18 * dt + old_jerk) * dt3
			     	+ old_acc) * dt2 + vel) * dt + pos;
		pred_vel = ((6 * k_over_18 + old_jerk) * dt2
			    	+ old_acc) * dt + vel;
		t_pred = t;
	    }
	}

	void flat_accumulate_acc_and_jerk(_dyn_ *, real);
	void flat_calculate_acc_and_jerk(_dyn_ *, real);
        void flat_set_first_timestep(real, real);
	void flat_update(const real, const real);
        bool flat_correct();
};

typedef vector (particle::*VMF_ptr)(void);     // vector member function pointer
typedef void (particle::*MF_ptr)(const vector &);     // member function pointer

typedef void (particle::*VF_ptr)(void);     // void member function pointer
typedef void (particle::*RF_ptr)(real);     // void member function
						    // pointer with real arg
typedef void (particle::*RRF_ptr)(real,real);     // void member function
						    // pointer with two real args
class particle_system
    {
    private:
	int nsize;
	particle * pb;

    public:
	int n;
	int nbh;
	real time;
	real eps2;
	vector pos;
	vector vel;
	real mass;
	system(){
	    n = 0;
	    nbh = 0;
	    time = 0;
	    pb = NULL;
	    }
 
	void read(istream & );
	void write(ostream & );
	void atos(istream & );
	void stoa(ostream & );
	void dump();
	void apply_vf(VF_ptr);
	void apply_vf(RF_ptr, real);
	void apply_vf(RRF_ptr, real, real);
	void plot(real time);
	real kinetic_energy();
	real energy();
	void show_energy();
	void calculate_gravity();
	void create_uniform_sphere(int nbody, real power_index, real r0);
	void evolve( real tend);
	void calculate_cmterms();
	particle * get_particle_pointer(){return pb;}
};

#endif













